# -*- coding: utf-8 -*-
# Default Python Libraries
from typing import Type

# AristaFlow REST Libraries
from af_execution_manager.api.arista_flow_service_api import AristaFlowServiceApi
from aristaflow.service_provider import ServiceProvider


class SseHelper(object):
    """
    Helper for using Server Sent Events.
    """

    _service_provider: ServiceProvider = None

    def __init__(self, service_provider: ServiceProvider):
        """
        Constructor
        """
        self._service_provider = service_provider

    def register(self, api_klass: Type):
        afs: AristaFlowServiceApi = self._service_provider.get_arista_flow_service_api(api_klass)
        afs.register_sse()



# coding: utf-8
# Default Python Libraries
from typing import List, Type

# AristaFlow REST Libraries
from aristaflow.service_provider import ServiceProvider


class RemoteIteratorHandler(object):
    """
    Utilities for handling remote iterators
    """

    _service_provider: ServiceProvider = None

    def __init__(self, service_provider: ServiceProvider):
        """
        Constructor
        """
        self._service_provider = service_provider

    def _consume(self, target: List, iterator_data, attrib_name: str, iter_api, iter_api_method):
        if iterator_data is None:
            return
        if getattr(iterator_data, attrib_name):
            target += getattr(iterator_data, attrib_name)
        else:
            # no data left, end recursion
            return
        if iterator_data.dropped:
            return

        next_iter = iter_api_method(iter_api, iterator_data.iterator_id)
        self._consume(target, next_iter, attrib_name, iter_api_method)

    def consume(
        self, iterator_data, attrib_name: str, iter_api_type: Type, iter_api_method=None
    ) -> List:
        iter_api = self._service_provider.get_service(iter_api_type)
        if iter_api_method is None:
            iter_api_method = getattr(iter_api_type, "get_next")
        target = []
        self._consume(target, iterator_data, attrib_name, iter_api, iter_api_method)
        return target



# -*- coding: utf-8 -*-
# Default Python Libraries
from datetime import datetime
from typing import List

# AristaFlow REST Libraries
from af_org_model_manager.api.policy_resolution_api import PolicyResolutionApi
from af_org_model_manager.api.qa_remote_iterator_api import QaRemoteIteratorApi
from af_worklist_manager.api.poss_abs_remote_iterator_api import PossAbsRemoteIteratorApi
from af_worklist_manager.api.worklist_update_manager_api import WorklistUpdateManagerApi
from af_worklist_manager.models.absence_data import AbsenceData
from af_worklist_manager.models.absence_information import AbsenceInformation
from af_worklist_manager.models.qualified_agent import QualifiedAgent
from aristaflow.abstract_service import AbstractService
from aristaflow.utils import OrgUtils


class ExtendedAbsenceInformation(object):
    """Absence information providing additional and typed data."""

    absence_information: AbsenceInformation = None
    absentee: QualifiedAgent = None
    is_absent_now = False
    substitute_summary = ""
    substitute_agents: List[QualifiedAgent] = []
    absent_from: datetime = None
    absent_until: datetime = None

    def __init__(
        self,
        absentee: QualifiedAgent,
        is_absent_now: bool,
        substitute_summary: str,
        substitute_agents: List[QualifiedAgent],
        ai: AbsenceInformation,
    ):
        self.is_absent_now = is_absent_now
        self.substitute_summary = substitute_summary
        self.substitute_agents = substitute_agents
        self.absentee = absentee
        if ai:
            self.absent_from = ai.from_date
            self.absent_until = ai.to_date
            self.absence_information = ai


class AbsenceService(AbstractService):
    """
    Helper methods for implementing absences and substitutions
    """

    def get_possible_absentees(self) -> List[QualifiedAgent]:
        """Returns the possible absentees"""
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        return self._rem_iter_handler.consume(
            wum.get_possible_absentees(),
            "agents",
            PossAbsRemoteIteratorApi,
            PossAbsRemoteIteratorApi.poss_abs_get_next,
        )

    def get_absence_information(
        self, *qas: List[QualifiedAgent]
    ) -> List[ExtendedAbsenceInformation]:
        """Returns the absence information for the given agents, or for all agents if no agent is provided"""
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        pol_res: PolicyResolutionApi = self._service_provider.get_service(PolicyResolutionApi)
        if qas is None or len(qas) == 0:
            qas = self.get_possible_absentees()

        res: List[ExtendedAbsenceInformation] = []
        for absentee in qas:
            ai: AbsenceInformation = wum.get_absence_information(body=absentee)
            is_absent_now = False
            substitute_summary = ""
            substitute_agents = []
            if ai is not None:
                # from date not in the future
                absence_started = ai.from_date < datetime.now()
                absence_lasting = ai.to_date == 0 or ai.to_date > datetime.now()
                is_absent_now = absence_started and absence_lasting
                if ai.substitution_rule:
                    substitute_agents = self._rem_iter_handler.consume(
                        pol_res.resolve_policy(org_policy=ai.substitution_rule),
                        "agents",
                        QaRemoteIteratorApi,
                    )
                substitute_summary = OrgUtils.summarize_qa_list(*substitute_agents)
            else:
                ai = {}
            res.append(
                ExtendedAbsenceInformation(
                    absentee, is_absent_now, substitute_summary, substitute_agents, ai
                )
            )
        return res

    def is_absent_now(self, absent_from: datetime = None, absent_until: datetime = None) -> bool:
        """Returns whether the specified absence range is considered as absent now."""
        absence_started = absent_from < datetime.now()
        absence_lasting = absent_until == 0 or absent_until > datetime.now()
        return absence_started and absence_lasting

    def create_substitutes_staff_assignment_rule(self, qas: List[QualifiedAgent]) -> str:
        """Creates a staff assignment rule selecting the given agents as substitutes for someone."""
        if len(qas) == 0:
            # not set FIXME check API
            return ""
        sar = ""
        for qa in qas:
            qa_sar = self.create_staff_assignment_rule_for_qa(qa)
            if sar == "":
                sar = qa_sar
            else:
                sar = sar + " OR " + qa_sar
        return sar

    def create_staff_assignment_rule_for_qa(self, qa: QualifiedAgent) -> str:
        """Creates a staff assignment rule selecting exactly the given agent."""
        return OrgUtils.build_staff_assignment_rule_for_agent(qa.agent_id, qa.org_pos_id)

    def set_absent(
        self,
        absentee: QualifiedAgent,
        substitutes: List[QualifiedAgent],
        absent_from: datetime = None,
        absent_until: datetime = None,
    ):
        """Sets the given agent as absent using the proved information"""
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        substitution_rule = self.create_substitutes_staff_assignment_rule(substitutes)
        absence_data = AbsenceData(absentee, absent_from, absent_until, substitution_rule)
        wum.set_absent(body=absence_data)

    def set_present(self, absentee: QualifiedAgent):
        """Sets the agent present"""
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        wum.set_present(body=absentee)

    def drop_absence(self, absentee: QualifiedAgent):
        """Drops any current or future absence"""
        wum: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
        wum.drop_absence(body=absentee)



# Default Python Libraries
import base64
from typing import Dict, List, Type, TypeVar

# AristaFlow REST Libraries
import af_execution_manager
from af_execution_manager.api.activity_execution_control_api import ActivityExecutionControlApi
from af_org_model_manager.api.global_security_manager_api import GlobalSecurityManagerApi
from af_org_model_manager.models.authentication_data import AuthenticationData
from af_org_model_manager.models.org_pos_spec import OrgPosSpec
from af_org_model_manager.models.client_session_details import ClientSessionDetails
from af_org_model_manager.models.qualified_agent import QualifiedAgent
from af_remote_html_runtime_manager.api.synchronous_activity_starting_api import (
    SynchronousActivityStartingApi,
)
from af_worklist_manager.models.af_activity_reference import AfActivityReference
from af_worklist_manager.models.worklist_item import WorklistItem
from af_runtime_service import SimpleSessionContext
from aristaflow.absence_service import AbsenceService
from aristaflow.delegation_service import DelegationService
from aristaflow.execution_history_service import ExecutionHistoryService
from aristaflow.image_renderer_service import ImageRendererService
from aristaflow.org_model_service import OrgModelService
from aristaflow.process_service import ProcessService
from aristaflow.service_provider import ServiceProvider
from aristaflow.activity_context import ActivityContext
from .abstract_service import AbstractService

from .activity_service import ActivityService
from .configuration import Configuration
from .remote_html_service import RemoteHtmlService, HtmlSignalHandler
from .worklist_service import WorklistService


T = TypeVar("T")
D = TypeVar("D")


class AristaFlowClientService(object):
    """Client Services for accessing the AristaFlow BPM platform"""

    # the user session id this client service belongs to
    __user_session: str = None
    # authentication
    __csd: ClientSessionDetails = None
    # the AristaFlow configuration
    __af_conf: Configuration = None
    __service_provider: ServiceProvider = None
    __worklist_service: WorklistService = None
    __process_service: ProcessService = None
    __delegation_service: DelegationService = None
    __absence_service: AbsenceService = None
    __execution_history_service: ExecutionHistoryService = None
    __image_renderer_service: ImageRendererService = None
    __org_model_service: OrgModelService = None
    __activity_service: ActivityService = None
    __remote_html_service: RemoteHtmlService = None
    __all_services: List[AbstractService] = []

    def __init__(
        self, configuration: Configuration, user_session: str, service_provider: ServiceProvider
    ):
        self.__af_conf = configuration
        self.__user_session = user_session
        self.__service_provider = service_provider

    def get_service(self, service_type: Type[T]) -> T:
        """
        Returns a service instance for the given service type, e.g.
        get_service(InstanceControlApi)
        @param service_type The class of the requested service.
        """
        return self.__service_provider.get_service(service_type)

    @property
    def client_session_details(self) -> ClientSessionDetails:
        """Returns the client session details of this service
        :return: The client session details of this service
        """
        return self.__csd

    @property
    def is_authenticated(self) -> bool:
        """Returns true, if this client service is already authenticated"""
        return self.__csd is not None

    def authenticate(self, username: str, password: str = None, org_pos_id: int = None):
        if self.__csd is not None:
            raise Exception("Already authenticated")

        auth_data: AuthenticationData = AuthenticationData(OrgPosSpec(user_name=username))

        if org_pos_id is not None:
            auth_data.org_pos_spec = OrgPosSpec(org_pos_id, username)
        psk = self.__af_conf.pre_shared_key
        method: str
        # if a password was provided, use it
        if password:
            auth_data.password = password
            method = "UTF-8_PASSWORD"
        # if PSK is configured, use that
        elif psk:
            # get the utf-8 bytes, encode them using base 64 and decode the resulting bytes using ASCII
            psk_encoded = base64.b64encode(bytes(psk, "UTF-8")).decode("ascii")
            auth_data.data = psk_encoded
            method = "SHARED_UTF-8_KEY"
        else:
            raise Exception("No authentication method left")

        gsm: GlobalSecurityManagerApi = self.get_service(GlobalSecurityManagerApi)

        # use a provided application name
        if self.__af_conf.application_name:
            if org_pos_id is None:
                # if an application name is provided, an org position ID must be used as well
                # get the org positions
                agents: List[QualifiedAgent] = gsm.pre_authenticate_method(auth_data, method)
                agent: QualifiedAgent
                # pick the single org position
                if len(agents) == 1:
                    agent = agents[0]
                # none: can't login
                elif len(agents) == 0:
                    raise Exception(
                        f"User does not have an org position {username} (supplied org position id: {org_pos_id})"
                    )
                else:
                    # use the first org position, except there is a agent_name/username match
                    agent = agents[0]
                    for a in agents:
                        if a.agent_name == username:
                            agent = a
                            break
                # set the org position for the actual authentication
                auth_data.org_pos_spec = OrgPosSpec(agent.org_pos_id, username)
            # use the application name
            auth_data.app_name = self.__af_conf.application_name

        csds: List[ClientSessionDetails] = gsm.authenticate_all_method(
            auth_data, method, self.__af_conf.caller_uri
        )

        csd: ClientSessionDetails
        if len(csds) == 1:
            csd = csds[0]
        elif len(csds) == 0:
            raise Exception(
                f"User does not have an org position {username} (supplied org position id: {org_pos_id})"
            )
        else:
            # pick the first as default
            csd = csds[0]
            # pick the one where username and org position name are the same
            for entry in csds:
                if entry.agent.agent.org_pos_name == entry.agent.agent.agent_name:
                    csd = entry
                    break

        self.__service_provider.authenticated(csd)
        self.__csd = csd

    def is_html_activity(self, item: WorklistItem) -> bool:
        return RemoteHtmlService.is_html_activity(item)

    @property
    def worklist_service(self):
        if self.__worklist_service is None:
            self.__worklist_service = WorklistService(self.__service_provider, self.__af_conf)
            self.__all_services.append(self.__worklist_service)
        return self.__worklist_service

    @property
    def process_service(self):
        if self.__process_service is None:
            self.__process_service = ProcessService(self.__service_provider, self.__af_conf)
            self.__all_services.append(self.__process_service)
        return self.__process_service

    @property
    def delegation_service(self):
        if self.__delegation_service is None:
            self.__delegation_service = DelegationService(self.__service_provider)
            self.__all_services.append(self.__delegation_service)
        return self.__delegation_service

    @property
    def absence_service(self):
        if self.__absence_service is None:
            self.__absence_service = AbsenceService(self.__service_provider)
            self.__all_services.append(self.__absence_service)
        return self.__absence_service

    @property
    def execution_history_service(self):
        if self.__execution_history_service is None:
            self.__execution_history_service = ExecutionHistoryService(self.__service_provider)
            self.__all_services.append(self.__execution_history_service)
        return self.__execution_history_service

    @property
    def image_renderer_service(self):
        if self.__image_renderer_service is None:
            self.__image_renderer_service = ImageRendererService(self.__service_provider)
            self.__all_services.append(self.__image_renderer_service)
        return self.__image_renderer_service

    @property
    def org_model_service(self):
        if self.__org_model_service is None:
            self.__org_model_service = OrgModelService(self.__service_provider)
            self.__all_services.append(self.__org_model_service)
        return self.__org_model_service

    @property
    def activity_service(self):
        if self.__activity_service is None:
            self.__activity_service = ActivityService(self.__service_provider, self.__af_conf)
            self.__all_services.append(self.__activity_service)
        return self.__activity_service

    @property
    def remote_html_service(self):
        if self.__remote_html_service is None:
            self.__remote_html_service = RemoteHtmlService(self.__service_provider)
            self.__all_services.append(self.__remote_html_service)
        return self.__remote_html_service

    def start_html_activity(self, item: WorklistItem, callback_url: str = None,
                            signal_handler: HtmlSignalHandler = None):
        """
        Starts the given HTML GUI worklist item using the Remote HTML Runtime Manager
        """
        return self.remote_html_service.start_html_activity(item, callback_url=callback_url,
                                                            signal_handler=signal_handler)

    def get_html_activity_starting(self) -> SynchronousActivityStartingApi:
        """
        Returns the Remote HTML Runtime Manager Synchronous Activity Starting, ensuring logon to the Runtime Manager
        """
        return self.remote_html_service.get_html_activity_starting()

    def reset_activity(self, item: WorklistItem):
        """Resets the given worklist item."""
        if item.state != "STARTED":
            # nothing to do
            return
        # TODO check for a "local" activity for a soft reset
        ar: AfActivityReference = item.act_ref
        ebp_ir = af_execution_manager.EbpInstanceReference(
            ar.type,
            ar.instance_id,
            ar.instance_log_id,
            ar.base_template_id,
            ar.node_id,
            ar.node_iteration,
            ar.execution_manager_uris,
            ar.runtime_manager_uris,
        )
        aec: ActivityExecutionControlApi = self.get_service(ActivityExecutionControlApi)
        aec.reset_to_prev_savepoint(body=ebp_ir, force=True)

    def deserialize(self, data, klass: Type[D]) -> D:
        """Deserialize data using the given class of the generated OpenAPI models."""
        return self.__service_provider.deserialize(data, klass)

    def serialize(self, obj) -> Dict:
        """Serialize REST model object"""
        return self.__service_provider.serialize(obj)

    def parse_ac_to_dict(self, ac: ActivityContext) -> dict:
        """
        Parses the context data required to recreate an ActivityContext
        to a json serializable dict
        """
        return {
            "token": ac.token,
            "ssc": self.serialize(ac.ssc)
        }

    def get_ac_from_dict(self,context_data: dict) -> ActivityContext:
        """
        Creates a new ActivityContext from a dict provided by parse_ac_to_dict
        """
        ac = ActivityContext()
        ac.token = context_data["token"]
        ac.ssc = self.deserialize(context_data["ssc"], SimpleSessionContext)
        return ac

    @property
    def autostart_timeout_seconds(self) -> int:
        """Wait time in seconds for auto start signals"""
        return self.__af_conf.autostart_timeout_seconds

    def disconnect(self):
        """ Disconnect all services"""
        for service in self.__all_services:
            try:
                service.disconnect()
            except Exception as e:
                print("Exception on disconnecting service", e)



# Default Python Libraries
from typing import Dict, List

# AristaFlow REST Libraries
from af_runtime_service import DataContext, ParameterValue, SimpleSessionContext


class ActivityContext(object):
    _ssc: SimpleSessionContext
    _input_parameters: Dict[str, ParameterValue]
    _output_parameters: Dict[str, ParameterValue]
    _token: str

    def __init__(self):
        self._input_parameters = {}
        self._output_parameters = {}

    @property
    def ssc(self):
        return self._ssc

    @ssc.setter
    def ssc(self, ssc: SimpleSessionContext):
        self._ssc = ssc
        self._map_parameters()

    @property
    def session_id(self) -> str:
        # session_id field does not exist in "IncompleteSSC" used by get_ssc
        return self._ssc.session_id if hasattr(self._ssc, 'session_id') else None

    def _map_parameters(self):
        """
        Builds the internal parameter value maps
        """
        dc: DataContext = self._ssc.data_context
        if dc.values is not None:
            for pv in dc.values:
                self._input_parameters[pv.name] = pv
        if dc.output_values is not None:
            for pv in dc.output_values:
                self._output_parameters[pv.name] = pv

    @property
    def input_parameters(self) -> List[ParameterValue]:
        """
        Returns the input parameter values for the data context, never returns None
        """
        return self._ssc.data_context.values or []

    @property
    def output_parameters(self) -> List[ParameterValue]:
        """
        Returns the output parameter values for the data context, never returns None
        """
        return self._ssc.data_context.output_values or []

    def get_input(self) -> Dict:
        """
        Returns all input parameters as dictionary
        """
        values = {}
        pv: ParameterValue
        for pv in self._ssc.data_context.values:
            values[pv.name] = pv.value
        return values

    def set_output(self, values: Dict):
        """
        Set the output parameters to the given values
        """
        for k in self.output_parameters:
            if k.name in values:
                self._output_parameters[k.name].value = values[k.name]

    @property
    def token(self) -> str:
        """ Returns the security token used to start / resume the activity """
        return self._token

    @token.setter
    def token(self, token: str):
        self._token = token



# AristaFlow REST Libraries
from af_process_image_renderer import RenderOptions
from af_process_image_renderer.api.process_image_renderer_api import ProcessImageRendererApi
from aristaflow.abstract_service import AbstractService


class ImageRendererService(AbstractService):

    def get_process_image_instance(self, inst_id):
        pir: ProcessImageRendererApi = self._service_provider.get_service(ProcessImageRendererApi)

        return pir.get_instance(RenderOptions(type='PNG', data_edges_visible=False, data_elements_visible=False,
                                                                              data_flow_lane_visible=False,
                                                                              decision_data_edges_visible=False,
                                                                              decision_data_elements_visible=False,
                                                                              system_data_elements_visible=False,
                                                                              hidden_nodes_visible=False,
                                                                              sync_edges_visible=True,
                                                                              system_data_edges_visible=False,
                                                                              adv_options={"GraphFigureProviderID":
                                                                                           "de.aristaflow.adept2.ui.processvisualisation.widgets.bpmn.BPMNSymbols"}), inst_id)



# -*- coding: utf-8 -*-
# Default Python Libraries
from typing import Dict, List, Union

# AristaFlow REST Libraries
from af_worklist_manager.api.del_rec_remote_iterator_api import DelRecRemoteIteratorApi
from af_worklist_manager.api.delegation_manager_api import DelegationManagerApi
from af_worklist_manager.models.delg_rec_with_comment import DelgRecWithComment
from af_worklist_manager.models.qa_initial_remote_iterator_data import QaInitialRemoteIteratorData
from af_worklist_manager.models.qa_remote_iterator_data import QaRemoteIteratorData
from af_worklist_manager.models.qualified_agent import QualifiedAgent
from af_worklist_manager.models.worklist_item import WorklistItem
from aristaflow.abstract_service import AbstractService


class DelegationService(AbstractService):
    """
    Helper methods for implementing delegation
    """

    def get_delegation_recipients(self, *items: WorklistItem) -> List[QualifiedAgent]:
        """Retrieves the possible delegation recipients for the given worklist items, ie.
        the intersection of all possible delegation recipients of all items.
        """
        recipient_set: Dict[str, QualifiedAgent] = None
        for item in items:
            recs = self._get_delegation_recipients(item)
            rec: QualifiedAgent = None
            if recipient_set is None:
                # initial iteration: add what is there
                recipient_set = {}
                for rec in recs:
                    recipient_set[f"{rec.agent_id}_{rec.org_pos_id}"] = rec
            else:
                # not the initial iteration: remove what doesn't exist
                keys_to_delete: List[str] = []
                for added_rec_key in recipient_set:
                    added_rec = recipient_set[added_rec_key]
                    found = False
                    for rec in recs:
                        if (
                            rec.agent_id == added_rec.agent_id
                            and rec.org_pos_id == added_rec.org_pos_id
                        ):
                            found = True
                            break
                    if not found:
                        keys_to_delete.append(added_rec_key)
                # remove from target set all entries which are not found in the delegation
                # recipients of the current item_id
                for key_to_delete in keys_to_delete:
                    del recipient_set[key_to_delete]

            # empty intersection, no need to continue
            if len(recipient_set) == 0:
                break

        return recipient_set.values()

    def _get_delegation_recipients(self, item: WorklistItem) -> List[QualifiedAgent]:
        """Retrieves the possible delegation recipients for the given worklist item"""
        dm: DelegationManagerApi = self._service_provider.get_service(DelegationManagerApi)
        drec: QaInitialRemoteIteratorData = dm.get_delegation_recipients(item.id)
        recipients: List[QualifiedAgent] = []
        self._iterate_delegation_recipients(recipients, drec)
        return recipients

    def _iterate_delegation_recipients(
        self,
        recipients: List[QualifiedAgent],
        cur_iter: Union[QaInitialRemoteIteratorData, QaRemoteIteratorData],
    ):
        """Reads up the given delegation recipients iterator"""
        if cur_iter is None:
            return
        if cur_iter.qas:
            recipients += cur_iter.qas
        else:
            return
        if cur_iter.closed:
            return

        api: DelRecRemoteIteratorApi = self._service_provider.get_service(DelRecRemoteIteratorApi)
        next_iter = api.del_rec_get_next(cur_iter.iterator_id)
        self._iterate_delegation_recipients(recipients, next_iter)

    def delegate(self, item: WorklistItem, comment: str, *args):
        """Delegates the given worklist item using the given comments to all QualifiedAgents supplied in *args
        :param List[QualifiedAgents] args: Delegation recipients
        """
        dm: DelegationManagerApi = self._service_provider.get_service(DelegationManagerApi)
        body: DelgRecWithComment = DelgRecWithComment(recipients=args, comment=comment)
        dm.delegate_work_item(body, item.id)
